#include <QFontMetrics>
#include <QAbstractItemView>
#include <QToolButton>
#include <QEvent>
#include <QHBoxLayout>
#include <QHoverEvent>

#include <klocalizedstring.h>

#include "pageitemdelegate.h"

PageItemDelegate::PageItemDelegate(QAbstractItemView *view, QObject *parent)
    : QStyledItemDelegate(parent)
{
    m_view = view;
    m_contextBar = new QWidget(m_view->viewport());
    m_contextBar->hide();

    QPalette palette = m_contextBar->palette();
    //QBrush brush(QColor(41, 128, 185, 128));
    QBrush brush(QColor(0, 0, 0, 128));
    brush.setStyle(Qt::SolidPattern);
    palette.setBrush(QPalette::Active, QPalette::Base, brush);
    palette.setBrush(QPalette::Inactive, QPalette::Base, brush);
    palette.setBrush(QPalette::Disabled, QPalette::Base, brush);
    m_contextBar->setPalette(palette);
    m_contextBar->setAutoFillBackground(true);

    initButtons();

    view->viewport()->installEventFilter(this);
    view->viewport()->setAttribute(Qt::WA_Hover);
}

void PageItemDelegate::paint(QPainter *painter, const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    QStyleOptionViewItem newOption = option;
    newOption.decorationPosition = QStyleOptionViewItem::Top;
    newOption.displayAlignment = Qt::AlignCenter;

    QStyledItemDelegate::paint(painter, newOption, index);
}

QSize PageItemDelegate::sizeHint(const QStyleOptionViewItem &option, const QModelIndex &index) const
{
    QFontMetrics fMetrics(option.font);
    //int textWidth = fMetrics.width(index.data().toString()) - 5;
    QSize size = QStyledItemDelegate::sizeHint(option, index);

    return QSize(m_view->width() - 5, size.height() + fMetrics.height());

    //return QSize(size.width() - textWidth - 20, size.height() + fMetrics.height());
}

bool PageItemDelegate::eventFilter(QObject *object, QEvent *event)
{
    //return QStyledItemDelegate::eventFilter(object, event);

    if (event->type() == QEvent::Destroy || event->type() == QEvent::ChildRemoved || !m_view)
        return QStyledItemDelegate::eventFilter(object, event);

    if (object == m_view->viewport()) {
        switch (event->type()) {
        case QEvent::ToolTip:
            return true;

        case QEvent::HoverMove:
        case QEvent::HoverLeave:
            hoverEventFilter(static_cast<QHoverEvent*>(event));
            return true;

        default:
            return false;
        }
    } else {
        return QStyledItemDelegate::eventFilter(object, event);
    }
}

void PageItemDelegate::setOcrEnabled(bool enabled)
{
    m_recognizeButton->setEnabled(enabled);
}

void PageItemDelegate::initButtons()
{
    QHBoxLayout *layout = new QHBoxLayout(m_contextBar);
    layout->setContentsMargins(2, 2, 2, 2);
    layout->setSpacing(1);

    layout->addSpacerItem(new QSpacerItem(1, 1, QSizePolicy::Expanding, QSizePolicy::Fixed));

    m_showFileButton = new QToolButton(m_contextBar);
    m_showFileButton->setIcon(QIcon::fromTheme(QStringLiteral("folder")));
    layout->addWidget(m_showFileButton);
    connect(m_showFileButton, &QToolButton::clicked, this, [this](){ emit showFileWithPage(m_indexUnderCursor.data(Qt::UserRole).toString()); });

    QToolButton *btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("document-save")));
    btn->setToolTip(i18n("Save"));
    layout->addWidget(btn);
    connect(btn, &QToolButton::clicked, this, [this](){ emit savePage(m_indexUnderCursor.row()); });

    m_recognizeButton = new QToolButton(m_contextBar);
    m_recognizeButton->setIcon(QIcon(QStringLiteral(":/ksane/ocr.png")));
    m_recognizeButton->setToolTip(i18n("Recognize"));
    layout->addWidget(m_recognizeButton);
    connect(m_recognizeButton, &QToolButton::clicked, this, [this](){ emit recognizePage(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("object-rotate-left")));
    layout->addWidget(btn);
    btn->setToolTip(i18n("Rotate left"));
    connect(btn, &QToolButton::clicked, this, [this](){ emit rotateLeftPage(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("object-rotate-right")));
    btn->setToolTip(i18n("Rotate right"));
    layout->addWidget(btn);
    connect(btn, &QToolButton::clicked, this, [this](){ emit rotateRightPage(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("object-flip-horizontal")));
    layout->addWidget(btn);
    btn->setToolTip(i18n("Horizontal mirror"));
    connect(btn, &QToolButton::clicked, this, [this](){ emit horizontalMirror(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("object-flip-vertical")));
    layout->addWidget(btn);
    btn->setToolTip(i18n("Vertical mirror"));
    connect(btn, &QToolButton::clicked, this, [this](){ emit verticalMirror(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("draw-arrow-up")));
    btn->setToolTip(i18n("Move up"));
    layout->addWidget(btn);
    connect(btn, &QToolButton::clicked, this, [this](){ emit moveUpPage(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("draw-arrow-down")));
    btn->setToolTip(i18n("Move down"));
    layout->addWidget(btn);
    connect(btn, &QToolButton::clicked, this, [this](){ emit moveDownPage(m_indexUnderCursor.row()); });

    btn = new QToolButton(m_contextBar);
    btn->setIcon(QIcon::fromTheme(QStringLiteral("edit-delete")));
    btn->setToolTip(i18n("Remove"));
    layout->addWidget(btn);
    connect(btn, &QToolButton::clicked, this, [this](){ emit removePage(m_indexUnderCursor.row()); });

    layout->addSpacerItem(new QSpacerItem(1, 1, QSizePolicy::Expanding, QSizePolicy::Fixed));
}

bool PageItemDelegate::hoverEventFilter(QHoverEvent *event)
{
    QModelIndex index = m_view->indexAt(event->pos());
    if (index != m_indexUnderCursor) {
        updateHoverUi(index);
    } else {
        m_view->update(m_indexUnderCursor);
    }
    return false;
}

void PageItemDelegate::updateHoverUi(const QModelIndex &index)
{
    QModelIndex oldIndex = m_indexUnderCursor;
    m_indexUnderCursor = index;
    m_view->update(oldIndex);

    if (m_indexUnderCursor.isValid()) {
        QIcon icon = index.data(Qt::DecorationRole).value<QIcon>();

        QSize size = icon.actualSize(QSize(256, 256));

        QRect rect = m_view->visualRect(m_indexUnderCursor);
        showContextBar(rect);
    } else {
        m_contextBar->hide();
    }
}

void PageItemDelegate::showContextBar(const QRect &rect)
{       
    const QString savePath = m_indexUnderCursor.data(Qt::UserRole).toString();

    if (!savePath.isEmpty())
        m_showFileButton->setToolTip(i18n("Show file %1", savePath));

    m_showFileButton->setVisible(!savePath.isEmpty());

    m_contextBar->move(rect.topLeft());
    m_contextBar->setFixedWidth(rect.width());
    m_contextBar->show();
}
