#include <QIcon>
#include <QMimeData>

#include <QDebug>

#include "pagesmodel.h"

PagesModel::PagesModel(QObject *parent)
    : QAbstractListModel(parent)
    , m_dropAction(Qt::MoveAction)
{
}

int PagesModel::rowCount(const QModelIndex &parent) const
{
    // For list models only the root node (an invalid parent) should return the list's size. For all
    // other (valid) parents, rowCount() should return 0 so that it does not become a tree model.
    if (parent.isValid())
        return 0;

    // FIXME: Implement me!
    return m_pages.count();
}

QVariant PagesModel::data(const QModelIndex &index, int role) const
{
    if (!index.isValid()) {
        return QVariant();
    }

    if (role == Qt::DecorationRole) {
        return QIcon(QPixmap::fromImage(m_pages.at(index.row())->image));
    }

    if (role == Qt::DisplayRole) {
        return QString::number(index.row() + 1);
    }

    if (role == Qt::UserRole) {
        return m_pages.at(index.row())->savePath;
    }

    // FIXME: Implement me!
    return QVariant();
}

Qt::ItemFlags PagesModel::flags(const QModelIndex &index) const
{
    if (!index.isValid())
        return Qt::NoItemFlags;

    return Qt::ItemIsSelectable | Qt::ItemIsEnabled | Qt::ItemIsDragEnabled | Qt::ItemIsDropEnabled;
}

bool PagesModel::moveRows(const QModelIndex &sourceParent, int sourceRow, int count, const QModelIndex &destinationParent, int destinationChild)
{
    if (count != 1 || destinationChild < 0)
        return false;

    beginMoveRows(sourceParent, sourceRow, sourceRow, destinationParent, destinationChild);
    //beginResetModel();
    m_pages.move(sourceRow, destinationChild);
    //endResetModel();
    endMoveRows();
    return true;
}

bool PagesModel::removeRows(int row, int count, const QModelIndex &parent)
{
    QModelIndexList indexes;

    for (int i = 0; i < count; ++i)
        indexes.append(index(row + i));

    removePages(indexes);
}

bool PagesModel::canDropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent) const
{
    Q_UNUSED(row);
    Q_UNUSED(parent);

    if (!data->hasFormat("application/page"))
        return false;

    if (action != Qt::MoveAction)
        return false;

    return true;
}

bool PagesModel::dropMimeData(const QMimeData *data, Qt::DropAction action, int row, int column, const QModelIndex &parent)
{
    if (!canDropMimeData(data, action, row, column, parent))
        return false;

    if (action == Qt::IgnoreAction)
        return true;

    QByteArray encodedData = data->data("application/page");
    QDataStream stream(&encodedData, QIODevice::ReadOnly);

    int srcRow = 0;
    stream >> srcRow;

    int dstRow = parent.row();

    beginResetModel();
    m_pages.move(srcRow, dstRow);
    endResetModel();

    return true;
}

QMimeData *PagesModel::mimeData(const QModelIndexList &indexes) const
{
    QMimeData *mimeData = new QMimeData;
    QByteArray encodedData;

    QDataStream stream(&encodedData, QIODevice::WriteOnly);

    for (const auto &index : indexes)
        if (index.isValid()) {
            int row = index.row();
            stream << row;
        }

    mimeData->setData("application/page", encodedData);
    return mimeData;
}

QStringList	PagesModel::mimeTypes() const
{
    QStringList types { QStringLiteral("application/page") };
    return types;
}

Qt::DropActions	PagesModel::supportedDropActions() const
{
    return m_dropAction;
}

void PagesModel::setMovePagesEnabled(bool enabled)
{
    if (enabled) {
        m_dropAction = Qt::MoveAction;
    } else {
        m_dropAction = Qt::IgnoreAction;
    }
}

void PagesModel::addPage(const KSanePagePointer &page)
{
    beginInsertRows(QModelIndex(), m_pages.size(), m_pages.size());
    m_pages.push_back(page);
    endInsertRows();
}

KSanePagePointer PagesModel::page(int pageNumber)
{
    if (pageNumber > -1 && pageNumber < m_pages.size()) {
        return m_pages.at(pageNumber);
    } else {
        return KSanePagePointer::create();
    }
}

void PagesModel::changePage(int pageNumber, const KSanePagePointer &page)
{
    if (pageNumber < m_pages.size()) {
        page->savePath.clear();
        m_pages[pageNumber] = page;
        emit dataChanged(index(pageNumber), index(pageNumber));
        emit layoutChanged();
    }
}

QList<KSanePagePointer> PagesModel::allPages()
{
    QList<KSanePagePointer> pages;

    for (int row = 0; row < rowCount(); ++row)
        pages.append(page(row));

    return pages;
}

void PagesModel::removePage(const QModelIndex &index)
{
    removePages(QModelIndexList() << index);
}

void PagesModel::removePages(const QModelIndexList& indexes)
{
    QList<int> pageNumbers;

    for (const auto &index : indexes)
        pageNumbers.append(index.row());

    auto moreThan = [](int a, int b) { return a > b; };

    std::sort(pageNumbers.begin(), pageNumbers.end(), moreThan);

    for (const auto &number : pageNumbers) {
        beginRemoveRows(QModelIndex(), number, number);
        m_pages.removeAt(number);
        endRemoveRows();
    }
}

void PagesModel::removeAllPages()
{
    beginRemoveRows(QModelIndex(), 0, m_pages.size() - 1);
    m_pages.clear();
    endRemoveRows();
}
