#include <QMessageBox>
#include <QProcess>

#include <klocalizedstring.h>

#include "pageswidget.h"
#include "pagesmodel.h"
#include "pageitemdelegate.h"
#include "ui_pageswidget.h"

PagesWidget::PagesWidget(QWidget *parent) :    
    m_ui(new Ui::PagesWidget)
{
    m_ui->setupUi(this);

    //setFixedWidth(230);
    setFixedWidth(300);

    m_pagesModel = new PagesModel;

    m_ui->m_pagesListView->setSelectionMode(QAbstractItemView::ExtendedSelection);
    m_ui->m_pagesListView->setDragEnabled(true);
    m_ui->m_pagesListView->setAcceptDrops(true);
    m_ui->m_pagesListView->setDropIndicatorShown(true);
    m_ui->m_pagesListView->setModel(m_pagesModel);
    m_ui->m_pagesListView->setContextMenuPolicy(Qt::NoContextMenu);

    QItemSelectionModel *selectionModel = m_ui->m_pagesListView->selectionModel();

    connect(selectionModel, &QItemSelectionModel::currentChanged, this, &PagesWidget::currentIndexChanged);
    connect(selectionModel, &QItemSelectionModel::selectionChanged, this, &PagesWidget::selectionChanged);

    initActions();
    initPageDelegate();
}

PagesWidget::~PagesWidget()
{
    delete m_ui;
}

void PagesWidget::addPage(const KSanePagePointer &page)
{
    m_pagesModel->addPage(page);

    QModelIndex idx = m_pagesModel->index(m_pagesModel->rowCount() - 1);

    m_ui->m_pagesListView->setCurrentIndex(idx);
}

QList<KSanePagePointer> PagesWidget::pages()
{
    QList<KSanePagePointer> pages;

    for (int i = 0; i < m_pagesModel->rowCount(); ++i)
        pages.append(m_pagesModel->page(i));

    return pages;
}

void PagesWidget::setOcrEnabled(bool enabled)
{
    m_ui->m_recognizeSelectedPagesAction->setEnabled(enabled);
    m_ui->m_recognizeAllPagesAction->setEnabled(enabled);

    m_pageDelegate->setOcrEnabled(enabled);
}

void PagesWidget::initActions()
{       
    m_ui->m_pagesListView->addAction(m_ui->m_saveSelectedPagesAction);
    connect(m_ui->m_saveSelectedPagesAction, &QAction::triggered, this, &PagesWidget::saveSelectedPages);

    m_ui->m_pagesListView->addAction(m_ui->m_saveAllPagesAction);
    connect(m_ui->m_saveAllPagesAction, &QAction::triggered, this, &PagesWidget::saveAllPages);

    m_ui->m_pagesListView->addAction(m_ui->m_recognizeSelectedPagesAction);
    connect(m_ui->m_recognizeSelectedPagesAction, &QAction::triggered, this, &PagesWidget::recognizeSelectedPages);

    m_ui->m_pagesListView->addAction(m_ui->m_recognizeAllPagesAction);
    connect(m_ui->m_recognizeAllPagesAction, &QAction::triggered, this, &PagesWidget::recognizeAllPages);

    m_ui->m_pagesListView->addAction(m_ui->m_rotatePagesLeftAction);
    connect(m_ui->m_rotatePagesLeftAction, &QAction::triggered, this, [this](){ rotateSelectedPages(-90); });

    m_ui->m_pagesListView->addAction(m_ui->m_rotatePagesRightAction);
    connect(m_ui->m_rotatePagesRightAction, &QAction::triggered, this, [this](){ rotateSelectedPages(90); });

    m_ui->m_pagesListView->addAction(m_ui->m_horizontalMirrorPagesAction);
    connect(m_ui->m_horizontalMirrorPagesAction, &QAction::triggered, this, [this](){ mirrorSelectedPages(true, false); });

    m_ui->m_pagesListView->addAction(m_ui->m_verticalMirrorPagesAction);
    connect(m_ui->m_verticalMirrorPagesAction, &QAction::triggered, this, [this](){ mirrorSelectedPages(false, true); });

    m_ui->m_pagesListView->addAction(m_ui->m_movePageUpAction);
    connect(m_ui->m_movePageUpAction, &QAction::triggered, this, &PagesWidget::moveUpCurrentPage);

    m_ui->m_pagesListView->addAction(m_ui->m_movePageDownAction);
    connect(m_ui->m_movePageDownAction, &QAction::triggered, this, &PagesWidget::moveDownCurrentPage);    

    m_ui->m_pagesListView->addAction(m_ui->m_removeSelectedPagesAction);
    connect(m_ui->m_removeSelectedPagesAction, &QAction::triggered, this, &PagesWidget::removeSelectedPages);

    m_ui->m_pagesListView->addAction(m_ui->m_removeAllPagesAction);    
    connect(m_ui->m_removeAllPagesAction, &QAction::triggered, this, &PagesWidget::removeAllPages);
}

void PagesWidget::initPageDelegate()
{
    m_pageDelegate = new PageItemDelegate(m_ui->m_pagesListView);

    m_ui->m_pagesListView->setItemDelegate(m_pageDelegate);

    connect(m_pageDelegate, &PageItemDelegate::savePage, this, &PagesWidget::savePage);
    connect(m_pageDelegate, &PageItemDelegate::recognizePage, this, &PagesWidget::recognizePage);
    connect(m_pageDelegate, &PageItemDelegate::rotateLeftPage, this, &PagesWidget::rotateLeftPage);
    connect(m_pageDelegate, &PageItemDelegate::rotateRightPage, this, &PagesWidget::rotateRightPage);
    connect(m_pageDelegate, &PageItemDelegate::horizontalMirror, this, &PagesWidget::horizontalMirror);
    connect(m_pageDelegate, &PageItemDelegate::verticalMirror, this, &PagesWidget::verticalMirror);
    connect(m_pageDelegate, &PageItemDelegate::moveUpPage, this, &PagesWidget::moveUpPage);
    connect(m_pageDelegate, &PageItemDelegate::moveDownPage, this, &PagesWidget::moveDownPage);
    connect(m_pageDelegate, &PageItemDelegate::removePage, this, &PagesWidget::removePage);

    connect(m_pageDelegate, &PageItemDelegate::showFileWithPage, this, [this](const QString &filePath){
        if (!filePath.isEmpty())
            QProcess::startDetached(QStringLiteral("fly-fm"), QStringList() << filePath);
    });
}

QList<KSanePagePointer> PagesWidget::pages(const QModelIndexList &indexes)
{
    QList<KSanePagePointer> pageList;

    for (const auto &index : indexes)
        pageList.append(m_pagesModel->page(index.row()));

    return pageList;
}

QList<KSanePagePointer> PagesWidget::selectedPages()
{
    QItemSelectionModel *selectionModel = m_ui->m_pagesListView->selectionModel();
    QModelIndexList indexList = selectionModel->selectedIndexes();
    QList<KSanePagePointer> pageList = pages(indexList);

    return pageList;
}

QList<int> PagesWidget::pageNumbers(const QModelIndexList &indexes)
{
    QList<int> numbers;

    for (const auto &index : indexes)
        numbers.append(index.row());

    return numbers;
}

QList<int> PagesWidget::selectedPageNumbres()
{
    QItemSelectionModel *selectionModel = m_ui->m_pagesListView->selectionModel();
    QModelIndexList indexList = selectionModel->selectedIndexes();
    QList<int> numbers = pageNumbers(indexList);

    return numbers;
}

void PagesWidget::rotatePage(int number, qreal angle)
{    
    KSanePagePointer page = m_pagesModel->page(number);

    QImage srcImage = page->image;

    QTransform transform;
    transform.rotate(angle);

    QImage dstImage = srcImage.transformed(transform, Qt::SmoothTransformation);

    page->image = dstImage;
    m_pagesModel->changePage(number, page);

    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (currentIndex.isValid() && currentIndex.row() == number)
        emit currentPageChanged(m_pagesModel->page(number));
}

void PagesWidget::rotateSelectedPages(qreal angle)
{
    const QList<int> numbers = selectedPageNumbres();

    for (const auto &number : numbers)
        rotatePage(number, angle);
}

void PagesWidget::mirrorPage(int number, bool horizontal, bool vertical)
{
    KSanePagePointer page = m_pagesModel->page(number);

    QImage srcImage = page->image;
    QImage dstImage = srcImage.mirrored(horizontal, vertical);

    page->image = dstImage;
    m_pagesModel->changePage(number, page);

    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (currentIndex.isValid() && currentIndex.row() == number)
        emit currentPageChanged(m_pagesModel->page(number));
}

void PagesWidget::mirrorSelectedPages(bool horizontal, bool vertical)
{
    const QList<int> numbers = selectedPageNumbres();

    for (const auto &number : numbers)
        mirrorPage(number, horizontal, vertical);
}

void PagesWidget::rotateCurrentPageLeft()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    rotatePage(currentIndex.row(), -90);
}

void PagesWidget::rotateCurrentPageRight()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    rotatePage(currentIndex.row(), 90);
}

void PagesWidget::horizontalMirrorCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    mirrorPage(currentIndex.row(), true, false);
}

void PagesWidget::verticalMirrorCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    mirrorPage(currentIndex.row(), false, true);
}

void PagesWidget::removeCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    removePage(currentIndex.row());
}

void PagesWidget::saveCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    savePage(currentIndex.row());
}

void PagesWidget::recognizeCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    recognizePage(currentIndex.row());
}

void PagesWidget::clipCurrentPage(const QRectF &rect)
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;   

    KSanePagePointer page = m_pagesModel->page(currentIndex.row());

    QImage srcImage = page->image;

    QRect transformRect;
    transformRect.setTop(rect.top() * srcImage.height());
    transformRect.setLeft(rect.left() * srcImage.width());
    transformRect.setBottom(rect.bottom() * srcImage.height());
    transformRect.setRight(rect.right() * srcImage.width());

    QImage dstImage = srcImage.copy(transformRect);

    int w = srcImage.width();
    int h = srcImage.height();

    page->image = dstImage;
    m_pagesModel->changePage(currentIndex.row(), page);

    emit currentPageChanged(m_pagesModel->page(currentIndex.row()));
}

void PagesWidget::rotateLeftPage(int number)
{
    rotatePage(number, -90);
}

void PagesWidget::rotateRightPage(int number)
{
    rotatePage(number, 90);
}

void PagesWidget::horizontalMirror(int number)
{
    mirrorPage(number, true, false);
}

void PagesWidget::verticalMirror(int number)
{
    mirrorPage(number, false, true);
}

void PagesWidget::moveUpPage(int number)
{
    if (number == 0)
        return;

    m_pagesModel->moveRow(QModelIndex(), number, QModelIndex(), number - 1);
}

void PagesWidget::moveUpCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    moveUpPage(currentIndex.row());
}

void PagesWidget::moveDownPage(int number)
{
    if (number == (m_pagesModel->rowCount() - 1))
        return;

    m_pagesModel->moveRow(QModelIndex(), number + 1, QModelIndex(), number);
}

void PagesWidget::moveDownCurrentPage()
{
    QModelIndex currentIndex = m_ui->m_pagesListView->currentIndex();

    if (!currentIndex.isValid())
        return;

    moveDownPage(currentIndex.row());
}

void PagesWidget::savePage(int number)
{
    KSanePagePointer page = m_pagesModel->page(number);
    emit savePages(QList<KSanePagePointer>() << page);
}

void PagesWidget::saveSelectedPages()
{        
    emit savePages(selectedPages());
}

void PagesWidget::saveAllPages()
{
    emit savePages(m_pagesModel->allPages());
}

void PagesWidget::recognizePage(int number)
{
    KSanePagePointer page = m_pagesModel->page(number);
    emit recognizePages(QList<KSanePagePointer>() << page);
}

void PagesWidget::recognizeSelectedPages()
{
    emit recognizePages(selectedPages());
}

void PagesWidget::recognizeAllPages()
{
    emit recognizePages(m_pagesModel->allPages());
}

void PagesWidget::removePage(int number)
{
    auto answer = QMessageBox::question(nullptr, i18n("Remove confirmation"), i18n("Do you want to remove page?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

    if (answer == QMessageBox::No)
        return;

    m_pagesModel->removePage(m_pagesModel->index(number));
}

void PagesWidget::removeSelectedPages()
{
    auto answer = QMessageBox::question(nullptr, i18n("Remove confirmation"), i18n("Do you want to remove selected pages?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

    if (answer == QMessageBox::No)
        return;

    QItemSelectionModel *selectionModel = m_ui->m_pagesListView->selectionModel();
    QModelIndexList indexes = selectionModel->selectedIndexes();

    m_pagesModel->removePages(indexes);

    m_ui->m_pagesListView->setCurrentIndex(m_pagesModel->index(0));
}

void PagesWidget::removeAllPages()
{
    auto answer = QMessageBox::question(nullptr, i18n("Remove confirmation"), i18n("Do you want to remove all pages?"), QMessageBox::Yes | QMessageBox::No, QMessageBox::No);

    if (answer == QMessageBox::No)
        return;

    m_pagesModel->removeAllPages();
}

void PagesWidget::currentIndexChanged(const QModelIndex &index)
{
    if (index.isValid()) {
        m_ui->m_pagesListView->setContextMenuPolicy(Qt::ActionsContextMenu);
    } else {
        m_ui->m_pagesListView->setContextMenuPolicy(Qt::NoContextMenu);
    }

    emit currentPageChanged(m_pagesModel->page(index.row()));
}

void PagesWidget::selectionChanged(const QItemSelection &selected, const QItemSelection &deselected)
{
    Q_UNUSED(selected);
    Q_UNUSED(deselected);

    QItemSelectionModel *selectionModel = m_ui->m_pagesListView->selectionModel();
    QModelIndexList indexesList = selectionModel->selectedIndexes();

    bool isOneItem = indexesList.size() < 2;

    m_ui->m_movePageUpAction->setEnabled(isOneItem);
    m_ui->m_movePageDownAction->setEnabled(isOneItem);

    m_pagesModel->setMovePagesEnabled(isOneItem);
}
