#include "pdfdocumentsaver.h"

#include <QMutex>
#include <QDebug>

#include <KSaneWidget>
#include <QUrl>

#include <QFileInfo>
#include <QPdfWriter>
#include <QPainter>

#include <QtMath>

//#include <skanlite_debug.h>

int pixelToPoint(int pixel, int dpi)
{
    return qFloor(double(pixel) / double(dpi) * 72. + 0.5);
}

struct PdfDocumentSaver::Private {
    bool   m_savedOk;
    QMutex m_runMutex;

    QUrl       m_url;
    QString    m_name;
    QByteArray m_data;
    int        m_width;
    int        m_height;
    int        m_bpl;
    int        m_dpi;
    int        m_format;    

    QImage     m_image;

    QPdfWriter *m_pdfWriter;
    QPainter   *m_painter;

    bool       m_firstPage;

    PdfDocumentSaver *q;

    bool savePage();

    void deletePointers();
};

// ------------------------------------------------------------------------
PdfDocumentSaver::PdfDocumentSaver(QObject *parent) : DocumentSaver(parent), d(new Private)
{
    d->q = this;
    d->m_pdfWriter = nullptr;
    d->m_painter   = nullptr;
}

// ------------------------------------------------------------------------
PdfDocumentSaver::~PdfDocumentSaver()
{
    d->deletePointers();
    delete d;
}

QString PdfDocumentSaver::format()
{
    return QStringLiteral("pdf");
}

QString PdfDocumentSaver::mimeType()
{
    return QStringLiteral("application/pdf");
}

QString PdfDocumentSaver::name()
{
    return d->m_name;
}

bool PdfDocumentSaver::newDocument(const QUrl &url, const QString &name, int dpi)
{
    if (!d->m_runMutex.tryLock())
        return false;

    d->m_url       = url;
    d->m_name      = name;
    d->m_firstPage = true;
    d->m_dpi       = dpi;
    d->m_pdfWriter = new QPdfWriter(d->m_name);

    d->m_runMutex.unlock();

    //qCDebug(SKANLITE_LOG) << "New PDF document";

    return true;
}

bool PdfDocumentSaver::saveDocument()
{
   // if (!d->m_runMutex.tryLock())
   //     return false;

    d->deletePointers();

    if (!d->m_name.isEmpty()) {
        emit documentSaved(d->m_name, d->m_savedOk);
        d->m_name.clear();
    }

    d->m_runMutex.unlock();

    //qCDebug(SKANLITE_LOG) << "Save PDF document";

    return true;
}

bool PdfDocumentSaver::savePage(const QByteArray &data, int width, int height, int bpl, int dpi, int format, const QString& fileFormat, int quality)
{
    if (!d->m_runMutex.tryLock() || !d->m_pdfWriter)
        return false;

    d->m_data   = data;
    d->m_width  = width;
    d->m_height = height;
    d->m_bpl    = bpl;
    d->m_dpi    = dpi;
    d->m_format = format;
    d->m_image  = QImage();

    start();
    return true;
}

bool PdfDocumentSaver::savePage(const QImage &page, int dpi)
{
    if (!d->m_runMutex.tryLock() || !d->m_pdfWriter)
        return false;

    d->m_dpi    = dpi;
    d->m_image  = page;

    start();
    return true;
}

void PdfDocumentSaver::run()
{
    d->m_savedOk = d->savePage();

    emit pageSaved(d->m_url, d->m_name, d->m_savedOk);

    d->m_runMutex.unlock();
}

bool PdfDocumentSaver::Private::savePage()
{
    //qCDebug(SKANLITE_LOG) << "Save PDF page";

    if (m_image.isNull()) {
        m_image = KSaneIface::KSaneWidget::toQImageSilent(m_data, m_width, m_height, m_bpl, m_dpi, (KSaneIface::KSaneWidget::ImageFormat) m_format);

        //qCDebug(SKANLITE_LOG) << QStringLiteral("Image: width=%1, height=%2, bpl=%3, dpi=%4, format=%5")
        //                                        .arg(m_image.width()).arg(m_image.height()).arg(m_bpl).arg(m_dpi).arg(m_format);
    } else {
        //qCDebug(SKANLITE_LOG) << QStringLiteral("Image: width=%1, height=%2, dpi=%3")
        //                                        .arg(m_image.width()).arg(m_image.height()).arg(m_dpi);
    }

    int h = pixelToPoint(m_image.height(), m_dpi);
    int w = pixelToPoint(m_image.width(), m_dpi);

    //qCDebug(SKANLITE_LOG) << QStringLiteral("PDF page: width=%1, height=%2")
    //                                        .arg(w).arg(h);

    QPageSize pageSize(QSize(w, h), QPageSize::Point, QString(), QPageSize::ExactMatch);

    m_pdfWriter->setPageSize(pageSize);
    m_pdfWriter->setPageMargins(QMargins(0, 0, 0, 0));
    m_pdfWriter->setResolution(m_dpi);

    if (!m_firstPage) {
        m_pdfWriter->newPage();
    } else {
        if (m_painter)
            delete m_painter;

        m_painter = new QPainter(m_pdfWriter);
    }

    m_firstPage = false;

    m_painter->drawImage(0, 0, m_image);

    return true;
}

void PdfDocumentSaver::Private::deletePointers()
{
    if (m_painter) {
        delete m_painter;
        m_painter = nullptr;
    }

    if (m_pdfWriter) {
        delete m_pdfWriter;
        m_pdfWriter = nullptr;
    }
}
